from dataclasses import dataclass
from pathlib import Path

from yaml import YAMLError

from cognite.neat._issues import NeatError


@dataclass(unsafe_hash=True)
class AuthorizationError(NeatError, RuntimeError):
    """Missing authorization for {action}: {reason}"""

    action: str
    reason: str


@dataclass(unsafe_hash=True)
class FileReadError(NeatError, RuntimeError):
    """Error when reading file, {filepath}: {reason}"""

    fix = "Is the {filepath} open in another program? Is the file corrupted?"
    filepath: Path
    reason: str


@dataclass(unsafe_hash=True)
class OxigraphStorageLockedError(NeatError, RuntimeError):
    """Oxigraph on-disk storage locked at the path {filepath}"""

    fix = "Delete LOCK file in the {filepath} folder to unlock the storage or use another storage path"
    filepath: Path


@dataclass(unsafe_hash=True)
class FileNotFoundNeatError(NeatError, FileNotFoundError):
    """File {filepath} not found"""

    fix = "Make sure to provide a valid file"
    filepath: Path


@dataclass(unsafe_hash=True)
class FileMissingRequiredFieldError(NeatError, ValueError):
    """Missing required {field_name} in {filepath}: {field}"""

    filepath: Path
    field_name: str
    field: str


@dataclass(unsafe_hash=True)
class NeatYamlError(NeatError, YAMLError):
    """Invalid YAML: {reason}"""

    extra = "Expected format: {expected_format}"
    fix = "Check if the file is a valid YAML file"

    reason: str
    expected_format: str | None = None


@dataclass(unsafe_hash=True)
class FileTypeUnexpectedError(NeatError, TypeError):
    """Unexpected file type: {filepath}. Expected format: {expected_format}"""

    filepath: Path
    expected_format: frozenset[str]


@dataclass(unsafe_hash=True)
class FileNotAFileError(NeatError, FileNotFoundError):
    """{filepath} is not a file"""

    fix = "Make sure to provide a valid file"
    filepath: Path


@dataclass(unsafe_hash=True)
class CDFMissingClientError(NeatError, RuntimeError):
    """CDF client is required: {reason}"""

    reason: str


@dataclass(unsafe_hash=True)
class CDFMissingResourcesError(NeatError, RuntimeError):
    """Following CDF resources are missing: {resources}"""

    resources: str



from abc import ABC
from dataclasses import dataclass
from typing import ClassVar, cast

from cognite.neat._issues import NeatError
from cognite.neat._utils.spreadsheet import SpreadsheetRead


@dataclass(unsafe_hash=True)
class SpreadsheetError(NeatError, ValueError, ABC):
    """In row {row}: {error}"""

    _name: ClassVar[str] = ""
    error: NeatError

    @classmethod
    def create(
        cls, location: tuple[int | str, ...], error: NeatError, spreadsheet: SpreadsheetRead | None = None
    ) -> "SpreadsheetError":
        spreadsheet_name = cast(str, location[0])
        if spreadsheet_name not in ERROR_CLS_BY_SPREADSHEET_NAME:
            # This happens for the metadata sheet, which are individual fields
            return MetadataValueError(error, field_name=spreadsheet_name)

        error_cls = ERROR_CLS_BY_SPREADSHEET_NAME[spreadsheet_name]
        row, column = cast(tuple[int, str], location[2:4])

        if spreadsheet:
            row = spreadsheet.adjusted_row_number(row)

        return error_cls(
            row=row,
            error=error,
            column=column,
        )


@dataclass(unsafe_hash=True)
class SpreadsheetListError(SpreadsheetError, ABC):
    """In row {row}, column '{column}': {error}"""

    row: int
    column: str


@dataclass(unsafe_hash=True)
class MetadataValueError(SpreadsheetError):
    """In field {field_name}: {error}"""

    _type: ClassVar[str] = "Metadata"
    field_name: str


@dataclass(unsafe_hash=True)
class ViewValueError(SpreadsheetListError):
    _name = "Views"


@dataclass(unsafe_hash=True)
class ContainerValueError(SpreadsheetListError):
    _name = "Containers"


@dataclass(unsafe_hash=True)
class PropertyValueError(SpreadsheetListError):
    _name = "Properties"


@dataclass(unsafe_hash=True)
class ClassValueError(SpreadsheetListError):
    _name = "Classes"


@dataclass(unsafe_hash=True)
class EnumValueError(SpreadsheetListError):
    _name = "Enum"


@dataclass(unsafe_hash=True)
class NodeValueError(SpreadsheetListError):
    _name = "Nodes"


ERROR_CLS_BY_SPREADSHEET_NAME = {cls_._name: cls_ for cls_ in SpreadsheetListError.__subclasses__()}

# Efficient way to set docstring for all classes
for _cls in ERROR_CLS_BY_SPREADSHEET_NAME.values():
    _cls.__doc__ = SpreadsheetListError.__doc__
del _cls



from datetime import datetime
from pathlib import Path

from cognite.client import data_modeling as dm
from rdflib import Graph, Namespace, URIRef

from cognite.neat._constants import get_default_prefixes_and_namespaces
from cognite.neat._issues import IssueList, MultiValueError
from cognite.neat._issues.errors import FileReadError
from cognite.neat._issues.errors._general import NeatValueError
from cognite.neat._rules._shared import ReadRules
from cognite.neat._rules.importers._base import BaseImporter
from cognite.neat._rules.models._base_rules import RoleTypes
from cognite.neat._rules.models.data_types import AnyURI
from cognite.neat._rules.models.entities import UnknownEntity
from cognite.neat._rules.models.information import InformationInputRules
from cognite.neat._store import NeatGraphStore
from cognite.neat._utils.rdf_ import get_namespace

DEFAULT_NON_EXISTING_NODE_TYPE = AnyURI()


DEFAULT_RDF_DATA_MODEL_ID = ("neat_space", "RDFDataModel", "rdf")


class BaseRDFImporter(BaseImporter[InformationInputRules]):
    """Baser RDF importers used for all rules importers that are using RDF as input.

    Args:
        issue_list: Issue list to store issues
        graph: Knowledge graph
        data_model_id: Data model id to be used for the imported rules
        space: CDF Space to be used for the imported rules
        language: Language for description and human readable entity names



    !!! note "Language"
        Language is provided as ISO 639-1 code. If not provided, English will be used as default.

    """

    def __init__(
        self,
        issue_list: IssueList,
        graph: Graph,
        data_model_id: dm.DataModelId | tuple[str, str, str],
        max_number_of_instance: int,
        non_existing_node_type: UnknownEntity | AnyURI,
        language: str,
        source_name: str = "Unknown",
    ) -> None:
        self.issue_list = issue_list
        self.graph = graph
        self.data_model_id = dm.DataModelId.load(data_model_id)
        if self.data_model_id.version is None:
            raise NeatValueError("Version is required when setting a Data Model ID")

        self.max_number_of_instance = max_number_of_instance
        self.non_existing_node_type = non_existing_node_type
        self.language = language
        self.source_name = source_name

    @classmethod
    def from_graph_store(
        cls,
        store: NeatGraphStore,
        data_model_id: (dm.DataModelId | tuple[str, str, str]) = DEFAULT_RDF_DATA_MODEL_ID,
        max_number_of_instance: int = -1,
        non_existing_node_type: UnknownEntity | AnyURI = DEFAULT_NON_EXISTING_NODE_TYPE,
        language: str = "en",
    ):
        return cls(
            IssueList(title=f"{cls.__name__} issues"),
            store.dataset,
            data_model_id=data_model_id,
            max_number_of_instance=max_number_of_instance,
            non_existing_node_type=non_existing_node_type,
            language=language,
        )

    @classmethod
    def from_file(
        cls,
        filepath: Path,
        data_model_id: (dm.DataModelId | tuple[str, str, str]) = DEFAULT_RDF_DATA_MODEL_ID,
        max_number_of_instance: int = -1,
        non_existing_node_type: UnknownEntity | AnyURI = DEFAULT_NON_EXISTING_NODE_TYPE,
        language: str = "en",
        source_name: str = "Unknown",
    ):
        issue_list = IssueList(title=f"{cls.__name__} issues")

        graph = Graph()
        try:
            graph.parse(filepath)
        except Exception as e:
            issue_list.append(FileReadError(filepath, str(e)))

        # bind key namespaces
        for prefix, namespace in get_default_prefixes_and_namespaces().items():
            graph.bind(prefix, namespace)

        return cls(
            issue_list,
            graph,
            data_model_id=data_model_id,
            max_number_of_instance=max_number_of_instance,
            non_existing_node_type=non_existing_node_type,
            language=language,
            source_name=source_name,
        )

    def to_rules(
        self,
    ) -> ReadRules[InformationInputRules]:
        """
        Creates `Rules` object from the data for target role.
        """
        if self.issue_list.has_errors:
            # In case there were errors during the import, the to_rules method will return None
            self.issue_list.trigger_warnings()
            raise MultiValueError(self.issue_list.errors)

        rules_dict = self._to_rules_components()

        rules = InformationInputRules.load(rules_dict)
        self.issue_list.trigger_warnings()
        return ReadRules(rules, {})

    def _to_rules_components(self) -> dict:
        raise NotImplementedError()

    @classmethod
    def _add_uri_namespace_to_prefixes(cls, URI: URIRef, prefixes: dict[str, Namespace]):
        """Add URI to prefixes dict if not already present

        Args:
            URI: URI from which namespace is being extracted
            prefixes: Dict of prefixes and namespaces
        """
        if Namespace(get_namespace(URI)) not in prefixes.values():
            prefixes[f"prefix_{len(prefixes) + 1}"] = Namespace(get_namespace(URI))

    @property
    def _metadata(self) -> dict:
        return {
            "role": RoleTypes.information,
            "space": self.data_model_id.space,
            "external_id": self.data_model_id.external_id,
            "version": self.data_model_id.version,
            "created": datetime.now().replace(microsecond=0),
            "updated": datetime.now().replace(microsecond=0),
            "name": None,
            "description": f"Data model imported using {type(self).__name__}",
            "creator": "Neat",
        }



import warnings
from abc import ABC
from collections import defaultdict
from typing import Any, ClassVar, Literal

from cognite.client import data_modeling as dm

from cognite.neat._client import NeatClient
from cognite.neat._issues.errors import CDFMissingClientError, NeatValueError, ResourceNotFoundError
from cognite.neat._issues.warnings import PropertyOverwritingWarning
from cognite.neat._rules.models import DMSRules, SheetList
from cognite.neat._rules.models.data_types import Enum
from cognite.neat._rules.models.dms import DMSContainer, DMSEnum, DMSProperty
from cognite.neat._rules.models.entities import ClassEntity, ContainerEntity, ViewEntity

from ._base import VerifiedRulesTransformer


class MapOntoTransformers(VerifiedRulesTransformer[DMSRules, DMSRules], ABC):
    """Base class for transformers that map one rule onto another."""

    ...


class MapOneToOne(MapOntoTransformers):
    """Takes transform data models and makes it into an extension of the reference data model.

    Note this transformer mutates the input rules.

    The argument view_extension_mapping is a dictionary where the keys are views of this data model,
    and each value is the view of the reference data model that the view should extend. For example:

    ```python
    view_extension_mapping = {"Pump": "Asset"}
    ```

    This would make the view "Pump" in this data model extend the view "Asset" in the reference data model.
    Note that all the keys in the dictionary must be external ids of views in this data model,
    and all the values must be external ids of views in the reference data model.

    Args:
        reference: The reference data model
        view_extension_mapping: A dictionary mapping views in this data model to views in the reference data model
        default_extension: The default view in the reference data model that views in this
            data model should extend if no mapping is provided.

    """

    def __init__(
        self, reference: DMSRules, view_extension_mapping: dict[str, str], default_extension: str | None = None
    ) -> None:
        self.reference = reference
        self.view_extension_mapping = view_extension_mapping
        self.default_extension = default_extension

    def transform(self, rules: DMSRules) -> DMSRules:
        solution: DMSRules = rules
        view_by_external_id = {view.view.external_id: view for view in solution.views}
        ref_view_by_external_id = {view.view.external_id: view for view in self.reference.views}

        if invalid_views := set(self.view_extension_mapping.keys()) - set(view_by_external_id.keys()):
            raise ValueError(f"Views are not in this dat model {invalid_views}")
        if invalid_views := set(self.view_extension_mapping.values()) - set(ref_view_by_external_id.keys()):
            raise ValueError(f"Views are not in the reference data model {invalid_views}")
        if self.default_extension and self.default_extension not in ref_view_by_external_id:
            raise ValueError(f"Default extension view not in the reference data model {self.default_extension}")

        properties_by_view_external_id: dict[str, dict[str, DMSProperty]] = defaultdict(dict)
        for prop in solution.properties:
            properties_by_view_external_id[prop.view.external_id][prop.view_property] = prop

        ref_properties_by_view_external_id: dict[str, dict[str, DMSProperty]] = defaultdict(dict)
        for prop in self.reference.properties:
            ref_properties_by_view_external_id[prop.view.external_id][prop.view_property] = prop

        for view_external_id, view in view_by_external_id.items():
            if view_external_id in self.view_extension_mapping:
                ref_external_id = self.view_extension_mapping[view_external_id]
            elif self.default_extension:
                ref_external_id = self.default_extension
            else:
                continue

            ref_view = ref_view_by_external_id[ref_external_id]
            shared_properties = set(properties_by_view_external_id[view_external_id].keys()) & set(
                ref_properties_by_view_external_id[ref_external_id].keys()
            )
            if shared_properties:
                if view.implements is None:
                    view.implements = [ref_view.view]
                elif isinstance(view.implements, list) and ref_view.view not in view.implements:
                    view.implements.append(ref_view.view)
            for prop_name in shared_properties:
                prop = properties_by_view_external_id[view_external_id][prop_name]
                ref_prop = ref_properties_by_view_external_id[ref_external_id][prop_name]
                if ref_prop.container and ref_prop.container_property:
                    prop.container = ref_prop.container
                    prop.container_property = ref_prop.container_property

        return solution


class RuleMapper(VerifiedRulesTransformer[DMSRules, DMSRules]):
    """Maps properties and classes using the given mapping.

    Args:
        mapping: The mapping to use represented as a DMSRules object.
        data_type_conflict: How to handle data type conflicts. The default is "overwrite".
            A data type conflicts occurs when the data type of a property in the mapping is different from the
            data type of the property in the input rules. If "overwrite" the data type in the input rules is overwritten
            with the data type in the mapping.
    """

    _mapping_fields: ClassVar[frozenset[str]] = frozenset(
        ["connection", "value_type", "nullable", "immutable", "is_list", "default", "index", "constraint"]
    )

    def __init__(self, mapping: DMSRules, data_type_conflict: Literal["overwrite"] = "overwrite") -> None:
        self.mapping = mapping
        self.data_type_conflict = data_type_conflict

    def transform(self, rules: DMSRules) -> DMSRules:
        if self.data_type_conflict != "overwrite":
            raise NeatValueError(f"Invalid data_type_conflict: {self.data_type_conflict}")
        input_rules = rules
        new_rules = input_rules.model_copy(deep=True)

        views_by_external_id = {view.view.external_id: view for view in new_rules.views}
        new_views: set[ViewEntity] = set()
        for mapping_view in self.mapping.views:
            if existing_view := views_by_external_id.get(mapping_view.view.external_id):
                existing_view.implements = mapping_view.implements
            else:
                # We need to add all the views in the mapping that are not in the input rules.
                # This is to ensure that all ValueTypes are present in the resulting rules.
                # For example, if a property is a direct relation to an Equipment view, we need to add
                # the Equipment view to the rules.
                new_rules.views.append(mapping_view)
                new_views.add(mapping_view.view)

        properties_by_view_property = {
            (prop.view.external_id, prop.view_property): prop for prop in new_rules.properties
        }
        existing_enum_collections = {item.collection for item in new_rules.enum or []}
        mapping_enums_by_collection: dict[ClassEntity, list[DMSEnum]] = defaultdict(list)
        for item in self.mapping.enum or []:
            mapping_enums_by_collection[item.collection].append(item)
        existing_containers = {container.container for container in new_rules.containers or []}
        mapping_containers_by_id = {container.container: container for container in self.mapping.containers or []}
        for mapping_prop in self.mapping.properties:
            if existing_prop := properties_by_view_property.get(
                (mapping_prop.view.external_id, mapping_prop.view_property)
            ):
                to_overwrite, conflicts = self._find_overwrites(existing_prop, mapping_prop)
                if conflicts and self.data_type_conflict == "overwrite":
                    warnings.warn(
                        PropertyOverwritingWarning(
                            existing_prop.view.as_id(), "view", existing_prop.view_property, tuple(conflicts)
                        ),
                        stacklevel=2,
                    )
                elif conflicts:
                    raise NeatValueError(
                        f"Conflicting properties for {existing_prop.view}.{existing_prop.view_property}: {conflicts}"
                    )

                for field_name, value in to_overwrite.items():
                    setattr(existing_prop, field_name, value)
                existing_prop.container = mapping_prop.container
                existing_prop.container_property = mapping_prop.container_property
            elif isinstance(mapping_prop.value_type, ViewEntity):
                # All connections must be included in the rules. This is to update the
                # ValueTypes of the implemented views.
                new_rules.properties.append(mapping_prop)
            elif "guid" in mapping_prop.view_property.casefold():
                # All guid properties are included. Theses are necessary to get an appropriate
                # filter on the resulting view.
                new_rules.properties.append(mapping_prop)
            else:
                # Skipping mapped properties that are not in the input rules.
                continue

            if (
                isinstance(mapping_prop.value_type, Enum)
                and mapping_prop.value_type.collection not in existing_enum_collections
            ):
                if not new_rules.enum:
                    new_rules.enum = SheetList[DMSEnum]([])
                new_rules.enum.extend(mapping_enums_by_collection[mapping_prop.value_type.collection])

            if (
                mapping_prop.container
                and mapping_prop.container not in existing_containers
                and (new_container := mapping_containers_by_id.get(mapping_prop.container))
            ):
                # Mapping can include new containers for GUID properties
                if not new_rules.containers:
                    new_rules.containers = SheetList[DMSContainer]([])
                new_rules.containers.append(new_container)

        return new_rules

    def _find_overwrites(self, prop: DMSProperty, mapping_prop: DMSProperty) -> tuple[dict[str, Any], list[str]]:
        """Finds the properties that need to be overwritten and returns them.

        In addition, conflicting properties are returned. Note that overwriting properties that are
        originally None is not considered a conflict. Thus, you can have properties to overwrite but no
        conflicts.

        Args:
            prop: The property to compare.
            mapping_prop: The property to compare against.

        Returns:
            A tuple with the properties to overwrite and the conflicting properties.

        """
        to_overwrite: dict[str, Any] = {}
        conflicts: list[str] = []
        for field_name in self._mapping_fields:
            mapping_value = getattr(mapping_prop, field_name)
            source_value = getattr(prop, field_name)
            if mapping_value != source_value:
                to_overwrite[field_name] = mapping_value
                if source_value is not None:
                    # These are used for warnings so we use the alias to make it more readable for the user
                    conflicts.append(mapping_prop.model_fields[field_name].alias or field_name)
        return to_overwrite, conflicts

    @property
    def description(self) -> str:
        return f"Mapping to {self.mapping.metadata.as_data_model_id()!r}."


class AsParentPropertyId(VerifiedRulesTransformer[DMSRules, DMSRules]):
    """Looks up all view properties that map to the same container property,
    and changes the child view property id to match the parent property id.
    """

    def __init__(self, client: NeatClient | None = None) -> None:
        self._client = client

    def transform(self, rules: DMSRules) -> DMSRules:
        input_rules = rules
        new_rules = input_rules.model_copy(deep=True)

        path_by_view = self._inheritance_path_by_view(new_rules)
        view_by_container_property = self._view_by_container_properties(new_rules)

        parent_view_property_by_container_property = self._get_parent_view_property_by_container_property(
            path_by_view, view_by_container_property
        )

        for prop in new_rules.properties:
            if prop.container and prop.container_property:
                if parent_name := parent_view_property_by_container_property.get(
                    (prop.container, prop.container_property)
                ):
                    prop.view_property = parent_name

        return new_rules

    # Todo: Move into Probe class. Note this means that the Probe class must take a NeatClient as an argument.
    def _inheritance_path_by_view(self, rules: DMSRules) -> dict[ViewEntity, list[ViewEntity]]:
        parents_by_view: dict[ViewEntity, list[ViewEntity]] = {view.view: view.implements or [] for view in rules.views}

        path_by_view: dict[ViewEntity, list[ViewEntity]] = {}
        for view in rules.views:
            path_by_view[view.view] = self._get_inheritance_path(
                view.view, parents_by_view, rules.metadata.as_data_model_id()
            )
        return path_by_view

    def _get_inheritance_path(
        self, view: ViewEntity, parents_by_view: dict[ViewEntity, list[ViewEntity]], data_model_id: dm.DataModelId
    ) -> list[ViewEntity]:
        if parents_by_view.get(view) == []:
            # We found the root.
            return [view]
        if view not in parents_by_view and self._client is not None:
            # Lookup the parent
            view_id = view.as_id()
            read_views = self._client.loaders.views.retrieve([view_id])
            if not read_views:
                # Warning? Should be caught by validation
                raise ResourceNotFoundError(view_id, "view", data_model_id, "data model")
            parent_view_latest = max(read_views, key=lambda view: view.created_time)
            parents_by_view[ViewEntity.from_id(parent_view_latest.as_id())] = [
                ViewEntity.from_id(grand_parent) for grand_parent in parent_view_latest.implements or []
            ]
        elif view not in parents_by_view:
            raise CDFMissingClientError(
                f"The data model {data_model_id} is referencing a view that is not in the data model."
                f"Please provide a client to lookup the view."
            )

        inheritance_path = [view]
        seen = {view}
        if view in parents_by_view:
            for parent in parents_by_view[view]:
                parent_path = self._get_inheritance_path(parent, parents_by_view, data_model_id)
                inheritance_path.extend([p for p in parent_path if p not in seen])
                seen.update(parent_path)
        return inheritance_path

    def _view_by_container_properties(
        self, rules: DMSRules
    ) -> dict[tuple[ContainerEntity, str], list[tuple[ViewEntity, str]]]:
        view_properties_by_container_properties: dict[tuple[ContainerEntity, str], list[tuple[ViewEntity, str]]] = (
            defaultdict(list)
        )
        view_with_properties: set[ViewEntity] = set()
        for prop in rules.properties:
            if not prop.container or not prop.container_property:
                continue
            view_properties_by_container_properties[(prop.container, prop.container_property)].append(
                (prop.view, prop.view_property)
            )
            view_with_properties.add(prop.view)

        # We need to look up all parent properties.
        to_lookup = {view.view.as_id() for view in rules.views if view.view not in view_with_properties}
        if to_lookup and self._client is None:
            raise CDFMissingClientError(
                f"Views {to_lookup} are not in the data model. Please provide a client to lookup the views."
            )
        elif to_lookup and self._client:
            read_views = self._client.loaders.views.retrieve(list(to_lookup), include_ancestor=True)
            write_views = [self._client.loaders.views.as_write(read_view) for read_view in read_views]
            # We use the write/request format of the views as the read/response format contains all properties
            # including ancestor properties. The goal is to find the property name used in the parent
            # and thus we cannot have that repeated in the child views.
            for write_view in write_views:
                view_id = write_view.as_id()
                view_entity = ViewEntity.from_id(view_id)

                for property_id, property_ in (write_view.properties or {}).items():
                    if not isinstance(property_, dm.MappedPropertyApply):
                        continue
                    container_entity = ContainerEntity.from_id(property_.container)
                    view_properties_by_container_properties[
                        (container_entity, property_.container_property_identifier)
                    ].append((view_entity, property_id))

        return view_properties_by_container_properties

    @staticmethod
    def _get_parent_view_property_by_container_property(
        path_by_view, view_by_container_properties: dict[tuple[ContainerEntity, str], list[tuple[ViewEntity, str]]]
    ) -> dict[tuple[ContainerEntity, str], str]:
        parent_name_by_container_property: dict[tuple[ContainerEntity, str], str] = {}
        for (container, container_property), view_properties in view_by_container_properties.items():
            if len(view_properties) == 1:
                continue
            # Shortest path is the parent
            _, prop_name = min(view_properties, key=lambda prop: len(path_by_view[prop[0]]))
            parent_name_by_container_property[(container, container_property)] = prop_name
        return parent_name_by_container_property

    @property
    def description(self) -> str:
        return "Renaming property names to parent name"



import json
import re
import sys
import typing
import urllib.parse
import warnings
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable, Sequence, Set
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Generic, TypeVar

from cognite.client import CogniteClient
from cognite.client.data_classes._base import WriteableCogniteResource
from cognite.client.exceptions import CogniteAPIError
from pydantic import AnyHttpUrl, ValidationError
from rdflib import RDF, XSD, Literal, Namespace, URIRef

from cognite.neat._constants import DEFAULT_NAMESPACE
from cognite.neat._graph.extractors._base import BaseExtractor
from cognite.neat._issues.errors import NeatValueError
from cognite.neat._issues.warnings import CDFAuthWarning, NeatValueWarning
from cognite.neat._shared import Triple
from cognite.neat._utils.auxiliary import string_to_ideal_type
from cognite.neat._utils.collection_ import iterate_progress_bar_if_above_config_threshold

T_CogniteResource = TypeVar("T_CogniteResource", bound=WriteableCogniteResource)

DEFAULT_SKIP_METADATA_VALUES = frozenset({"nan", "null", "none", ""})

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from backports.strenum import StrEnum


class InstanceIdPrefix(StrEnum):
    asset = "Asset_"
    label = "Label_"
    relationship = "Relationship_"
    sequence = "Sequence_"
    file = "File_"
    time_series = "TimeSeries_"
    event = "Event_"
    data_set = "DataSet_"

    @classmethod
    def from_str(cls, raw: str) -> "InstanceIdPrefix":
        raw = raw.title() + "_"
        if raw == "Timeseries_":
            return cls.time_series
        else:
            return cls(raw)


class ClassicCDFBaseExtractor(BaseExtractor, ABC, Generic[T_CogniteResource]):
    """This is the Base Extractor for all classic CDF resources.

    A classic resource is recognized in that it has a metadata attribute of type dict[str, str].

    Args:
        items (Iterable[T_CogniteResource]): An iterable of classic resource.
        namespace (Namespace, optional): The namespace to use. Defaults to DEFAULT_NAMESPACE.
        total (int, optional): The total number of items to load. If passed, you will get a progress bar if rich
            is installed. Defaults to None.
        limit (int, optional): The maximal number of items to load. Defaults to None. This is typically used for
            testing setup of the extractor. For example, if you are extracting 100 000 assets, you might want to
            limit the extraction to 1000 assets to test the setup.
        unpack_metadata (bool, optional): Whether to unpack metadata. Defaults to False, which yields the metadata as
            a JSON string.
        skip_metadata_values (set[str] | frozenset[str] | None, optional): If you are unpacking metadata, then
           values in this set will be skipped.
        camel_case (bool, optional): Whether to use camelCase instead of snake_case for property names.
            Defaults to True.
        as_write (bool, optional): Whether to use the write/request format of the items. Defaults to False.
        prefix (str, optional): A prefix to add to the rdf type. Defaults to None.
        identifier (Literal["id", "externalId"], optional): The identifier to use. Defaults to "id".
    """

    _default_rdf_type: str
    _instance_id_prefix: str
    _SPACE_PATTERN = re.compile(r"\s+")

    def __init__(
        self,
        items: Iterable[T_CogniteResource],
        namespace: Namespace | None = None,
        total: int | None = None,
        limit: int | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
        camel_case: bool = True,
        as_write: bool = False,
        prefix: str | None = None,
        identifier: typing.Literal["id", "externalId"] = "id",
    ):
        self.namespace = namespace or DEFAULT_NAMESPACE
        self.items = items
        self.total = total
        self.limit = min(limit, total) if limit and total else limit
        self.unpack_metadata = unpack_metadata
        self.skip_metadata_values = skip_metadata_values
        self.camel_case = camel_case
        self.as_write = as_write
        self.prefix = prefix
        self.identifier = identifier
        # If identifier=externalId, we need to keep track of the external ids
        # and use them in linking of Files, Sequences, TimeSeries, and Events.
        self.asset_external_ids_by_id: dict[int, str] = {}
        self.lookup_dataset_external_id: Callable[[int], str] | None = None
        # Used by the ClassicGraphExtractor to log URIRefs
        self._log_urirefs = False
        self._uriref_by_external_id: dict[str, URIRef] = {}

    def extract(self) -> Iterable[Triple]:
        """Extracts an asset with the given asset_id."""
        from ._assets import AssetsExtractor

        if self.total is not None and self.total > 0:
            to_iterate = iterate_progress_bar_if_above_config_threshold(
                self.items, self.total, f"Extracting {type(self).__name__.removesuffix('Extractor')}"
            )
        else:
            to_iterate = self.items
        if self.identifier == "externalId" and isinstance(self, AssetsExtractor):
            to_iterate = self._store_asset_external_ids(to_iterate)  # type: ignore[attr-defined]

        for no, asset in enumerate(to_iterate):
            yield from self._item2triples(asset)
            if self.limit and no >= self.limit:
                break

    def _store_asset_external_ids(self, items: Iterable[T_CogniteResource]) -> Iterable[T_CogniteResource]:
        for item in items:
            if hasattr(item, "id") and hasattr(item, "external_id"):
                self.asset_external_ids_by_id[item.id] = item.external_id
            yield item

    def _item2triples(self, item: T_CogniteResource) -> list[Triple]:
        if self.identifier == "id":
            id_value: str | None
            if hasattr(item, "id"):
                id_value = str(item.id)
            else:
                id_value = self._fallback_id(item)
            if id_value is None:
                return []
            id_suffix = id_value
        elif self.identifier == "externalId":
            if not hasattr(item, "external_id"):
                return []
            id_suffix = self._external_id_as_uri_suffix(item.external_id)
        else:
            raise NeatValueError(f"Unknown identifier {self.identifier}")

        id_ = self.namespace[f"{self._instance_id_prefix}{id_suffix}"]
        if self._log_urirefs and hasattr(item, "external_id"):
            self._uriref_by_external_id[item.external_id] = id_

        type_ = self._get_rdf_type()

        # Set rdf type
        triples: list[Triple] = [(id_, RDF.type, self.namespace[type_])]
        if self.as_write:
            item = item.as_write()
        dumped = item.dump(self.camel_case)
        dumped.pop("id", None)
        # We have parentId so we don't need parentExternalId
        dumped.pop("parentExternalId", None)
        if "metadata" in dumped:
            triples.extend(self._metadata_to_triples(id_, dumped.pop("metadata")))

        triples.extend(self._item2triples_special_cases(id_, dumped))

        for key, value in dumped.items():
            if value is None or value == []:
                continue
            values = value if isinstance(value, Sequence) and not isinstance(value, str) else [value]
            for raw in values:
                triples.append((id_, self.namespace[key], self._as_object(raw, key)))
        return triples

    def _item2triples_special_cases(self, id_: URIRef, dumped: dict[str, Any]) -> list[Triple]:
        """This can be overridden to handle special cases for the item."""
        return []

    @classmethod
    def _external_id_as_uri_suffix(cls, external_id: str | None) -> str:
        if external_id == "":
            warnings.warn(NeatValueWarning(f"Empty external id in {cls._default_rdf_type}"), stacklevel=2)
            return "empty"
        elif external_id == "\x00":
            warnings.warn(NeatValueWarning(f"Null external id in {cls._default_rdf_type}"), stacklevel=2)
            return "null"
        elif external_id is None:
            warnings.warn(NeatValueWarning(f"None external id in {cls._default_rdf_type}"), stacklevel=2)
            return "None"
        # The external ID needs to pass the ^[^\\x00]{1,256}$ regex for the DMS API.
        # In addition, neat internals requires the external ID to be a valid URI.
        return urllib.parse.quote(external_id)

    def _fallback_id(self, item: T_CogniteResource) -> str | None:
        raise AttributeError(
            f"Item of type {type(item)} does not have an id attribute. "
            "Please implement the _fallback_id method in the extractor."
        )

    def _metadata_to_triples(self, id_: URIRef, metadata: dict[str, str]) -> Iterable[Triple]:
        if self.unpack_metadata:
            for key, value in metadata.items():
                if value and (self.skip_metadata_values is None or value.casefold() not in self.skip_metadata_values):
                    yield (
                        id_,
                        self.namespace[urllib.parse.quote(key)],
                        Literal(string_to_ideal_type(value)),
                    )
        else:
            yield id_, self.namespace.metadata, Literal(json.dumps(metadata), datatype=XSD._NS["json"])

    def _get_rdf_type(self) -> str:
        type_ = self._default_rdf_type
        if self.prefix:
            type_ = f"{self.prefix}{type_}"
        return self._SPACE_PATTERN.sub("_", type_)

    def _as_object(self, raw: Any, key: str) -> Literal | URIRef:
        """Return properly formatted object part of s-p-o triple"""
        if key in {"data_set_id", "dataSetId"}:
            if self.identifier == "externalId" and self.lookup_dataset_external_id:
                try:
                    data_set_external_id = self.lookup_dataset_external_id(raw)
                except KeyError:
                    return Literal("Unknown data set")
                else:
                    return self.namespace[
                        f"{InstanceIdPrefix.data_set}{self._external_id_as_uri_suffix(data_set_external_id)}"
                    ]
            else:
                return self.namespace[f"{InstanceIdPrefix.data_set}{raw}"]
        elif key in {"assetId", "asset_id", "assetIds", "asset_ids", "parentId", "rootId", "parent_id", "root_id"}:
            if self.identifier == "id":
                return self.namespace[f"{InstanceIdPrefix.asset}{raw}"]
            else:
                try:
                    asset_external_id = self._external_id_as_uri_suffix(self.asset_external_ids_by_id[raw])
                except KeyError:
                    return Literal("Unknown asset", datatype=XSD.string)
                else:
                    return self.namespace[f"{InstanceIdPrefix.asset}{asset_external_id}"]
        elif key in {
            "startTime",
            "endTime",
            "createdTime",
            "lastUpdatedTime",
            "start_time",
            "end_time",
            "created_time",
            "last_updated_time",
        } and isinstance(raw, int):
            try:
                return Literal(datetime.fromtimestamp(raw / 1000, timezone.utc), datatype=XSD.dateTime)
            except (OSError, ValueError) as e:
                warnings.warn(NeatValueWarning(f"Failed to convert timestamp {raw} to datetime: {e!s}"), stacklevel=2)
                return Literal(raw)
        elif key == "labels":
            from ._labels import LabelsExtractor

            return self.namespace[f"{InstanceIdPrefix.label}{LabelsExtractor._label_id(raw)}"]
        elif key in {"sourceType", "targetType", "source_type", "target_type"} and isinstance(raw, str):
            # Relationship types. Titled so they can be looked up.
            return self.namespace[raw.title()]
        elif key in {"unit_external_id", "unitExternalId"}:
            try:
                return URIRef(str(AnyHttpUrl(raw)))
            except ValidationError:
                ...
        return Literal(raw)

    @classmethod
    def from_dataset(
        cls,
        client: CogniteClient,
        data_set_external_id: str,
        namespace: Namespace | None = None,
        limit: int | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
        camel_case: bool = True,
        as_write: bool = False,
        prefix: str | None = None,
        identifier: typing.Literal["id", "externalId"] = "id",
    ):
        total, items = cls._handle_no_access(lambda: cls._from_dataset(client, data_set_external_id))
        return cls(
            items,
            namespace,
            total,
            limit,
            unpack_metadata,
            skip_metadata_values,
            camel_case,
            as_write,
            prefix,
            identifier,
        )

    @classmethod
    @abstractmethod
    def _from_dataset(
        cls, client: CogniteClient, data_set_external_id: str
    ) -> tuple[int | None, Iterable[T_CogniteResource]]:
        raise NotImplementedError

    @classmethod
    def from_hierarchy(
        cls,
        client: CogniteClient,
        root_asset_external_id: str,
        namespace: Namespace | None = None,
        limit: int | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
        camel_case: bool = True,
        as_write: bool = False,
        prefix: str | None = None,
        identifier: typing.Literal["id", "externalId"] = "id",
    ):
        total, items = cls._handle_no_access(lambda: cls._from_hierarchy(client, root_asset_external_id))
        return cls(
            items,
            namespace,
            total,
            limit,
            unpack_metadata,
            skip_metadata_values,
            camel_case,
            as_write,
            prefix,
            identifier,
        )

    @classmethod
    @abstractmethod
    def _from_hierarchy(
        cls, client: CogniteClient, root_asset_external_id: str
    ) -> tuple[int | None, Iterable[T_CogniteResource]]:
        raise NotImplementedError

    @classmethod
    def from_file(
        cls,
        file_path: str | Path,
        namespace: Namespace | None = None,
        limit: int | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
        camel_case: bool = True,
        as_write: bool = False,
        prefix: str | None = None,
        identifier: typing.Literal["id", "externalId"] = "id",
    ):
        total, items = cls._from_file(file_path)
        return cls(
            items,
            namespace,
            total,
            limit,
            unpack_metadata,
            skip_metadata_values,
            camel_case,
            as_write,
            prefix,
            identifier,
        )

    @classmethod
    @abstractmethod
    def _from_file(cls, file_path: str | Path) -> tuple[int | None, Iterable[T_CogniteResource]]:
        raise NotImplementedError

    @classmethod
    def _handle_no_access(
        cls, action: Callable[[], tuple[int | None, Iterable[T_CogniteResource]]]
    ) -> tuple[int | None, Iterable[T_CogniteResource]]:
        try:
            return action()
        except CogniteAPIError as e:
            if e.code == 403:
                warnings.warn(
                    CDFAuthWarning(f"extract {cls.__name__.removesuffix('Extractor').casefold()}", str(e)), stacklevel=2
                )
                return 0, []
            else:
                raise e



import tempfile
from abc import ABC, abstractmethod
from collections.abc import Iterable
from io import StringIO
from pathlib import Path
from typing import IO, Any, TextIO
from urllib.parse import urlparse

import requests

from cognite.neat._issues.errors import NeatTypeError, NeatValueError


class NeatReader(ABC):
    @classmethod
    def create(cls, io: Any) -> "NeatReader":
        if isinstance(io, str):
            url = urlparse(io)
            if url.scheme == "https" and url.netloc.endswith("github.com"):
                return GitHubReader(io)
            elif url.scheme == "https":
                return HttpFileReader(io, url.path)

        if isinstance(io, str | Path):
            return PathReader(Path(io))
        raise NeatTypeError(f"Unsupported type: {type(io)}")

    @property
    def name(self) -> str:
        return str(self)

    @abstractmethod
    def read_text(self) -> str:
        """Read the buffer as a string"""
        raise NotImplementedError()

    @abstractmethod
    def read_bytes(self) -> bytes:
        """Read the buffer as bytes"""
        raise NotImplementedError()

    @abstractmethod
    def size(self) -> int:
        """Size of the buffer in bytes"""
        raise NotImplementedError()

    @abstractmethod
    def iterate(self, chunk_size: int) -> Iterable[str]:
        """Iterate over the buffer in chunks

        Args:
            chunk_size: Size of each chunk in bytes
        """
        raise NotImplementedError()

    @abstractmethod
    def __enter__(self) -> IO:
        raise NotImplementedError()

    @abstractmethod
    def __str__(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def exists(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def materialize_path(self) -> Path:
        raise NotImplementedError


class PathReader(NeatReader):
    def __init__(self, path: Path):
        self.path = path
        self._io: TextIO | None = None

    @property
    def name(self) -> str:
        return self.path.name

    def read_text(self) -> str:
        return self.path.read_text()

    def read_bytes(self) -> bytes:
        return self.path.read_bytes()

    def size(self) -> int:
        return self.path.stat().st_size

    def iterate(self, chunk_size: int) -> Iterable[str]:
        with self.path.open(mode="r") as f:
            while chunk := f.read(chunk_size):
                yield chunk

    def __enter__(self) -> TextIO:
        file = self.path.open(mode="r")
        self._io = file
        return file

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        if self._io:
            self._io.close()

    def __str__(self) -> str:
        return self.path.as_posix()

    def exists(self) -> bool:
        return self.path.exists()

    def materialize_path(self) -> Path:
        return self.path


class HttpFileReader(NeatReader):
    def __init__(self, url: str, path: str):
        self._url = url
        self.path = path

    @property
    def name(self) -> str:
        if "/" in self.path:
            return self.path.rsplit("/", maxsplit=1)[-1]
        return self.path

    def read_text(self) -> str:
        response = requests.get(self._url)
        response.raise_for_status()
        return response.text

    def read_bytes(self) -> bytes:
        response = requests.get(self._url)
        response.raise_for_status()
        return response.content

    def size(self) -> int:
        response = requests.head(self._url)
        response.raise_for_status()
        return int(response.headers["Content-Length"])

    def iterate(self, chunk_size: int) -> Iterable[str]:
        with requests.get(self._url, stream=True) as response:
            response.raise_for_status()
            for chunk in response.iter_content(chunk_size):
                yield chunk.decode("utf-8")

    def __enter__(self) -> IO:
        return StringIO(self.read_text())

    def __str__(self) -> str:
        return self._url

    def exists(self) -> bool:
        response = requests.head(self._url)
        return 200 <= response.status_code < 400

    def materialize_path(self) -> Path:
        path = Path(tempfile.gettempdir()).resolve() / "neat" / self.name
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_bytes(self.read_bytes())
        return path


class GitHubReader(HttpFileReader):
    raw_url = "https://raw.githubusercontent.com/"

    def __init__(self, raw: str):
        self.raw = raw
        self.repo, path = self._parse_url(raw)
        super().__init__(f"{self.raw_url}{self.repo}/main/{path}", path)

    @staticmethod
    def _parse_url(url: str) -> tuple[str, str]:
        parsed = urlparse(url)
        if parsed.scheme != "https":
            raise NeatValueError(f"Unsupported scheme: {parsed.scheme}")

        path = parsed.path.lstrip("/")
        if parsed.netloc == "github.com":
            repo, path = path.split("/blob/main/", maxsplit=1)
            return repo, path

        elif parsed.netloc == "api.github.com":
            repo, path = path.removeprefix("repos/").split("/contents/", maxsplit=1)
            return repo, path

        elif parsed.netloc == "raw.githubusercontent.com":
            repo, path = path.split("/main/", maxsplit=1)
            return repo, path

        raise NeatValueError(f"Unsupported netloc: {parsed.netloc}")

    def __str__(self) -> str:
        return self.raw



from collections.abc import Generator
from xml.etree.ElementTree import Element


def iterate_tree(node: Element, skip_root: bool = False) -> Generator:
    """Iterate over all elements in an XML tree.

    Args:
        node: XML tree to iterate over.

    Returns:
        Generator of XML elements.
    """
    if not skip_root:
        yield node
    for child in node:
        yield from iterate_tree(child)


def get_children(
    element: Element,
    child_tag: str,
    include_nested_children: bool = False,
    ignore_namespace: bool = False,
    no_children: int = -1,
) -> Element | list[Element]:
    """Get direct children of an XML element.

    Args:
        element: XML element to get children from.
        child_tag: Tag of the children to get.
        include_nested_children: bool to decide if only direct child elements should be extracted, or if all child
        elements (including nested ones) should be returned.
        ignore_namespace: bool that decides if wildcard * should be used to ignore namespace of children elements tag
        no_children: Max number of children to get. Defaults to -1 (all).

    Returns:
        List of XML elements if no_children > 1, otherwise XML element.
    """
    search_string = ""
    if ignore_namespace:
        if include_nested_children:
            search_string = f".//{{*}}{child_tag}"
        else:
            search_string = f".{{*}}{child_tag}"
    elif not ignore_namespace:
        if include_nested_children:
            search_string = f".//{child_tag}"
        else:
            search_string = f".{child_tag}"
    children = element.findall(search_string)
    return children[:no_children] if no_children > 0 else children
